from motor.motor_asyncio import AsyncIOMotorClient
from beanie import init_beanie
from beanie.odm.documents import DocType

from kmx.flows.db import DbFlowRun, DbFlowTaskRun

class BeanieDbInitilizer:

    def __init__(self, mongo_url: str, db_name: str, document_models: list[DocType]):
        self.mongo_url = mongo_url
        self.db_name = db_name
        self.document_models = document_models

    async def init_db(self):
        """ """
        mongo_client = AsyncIOMotorClient(self.mongo_url, tz_aware=True)

        document_models = []
        document_models.extend(self.document_models)
        document_models.extend([DbFlowRun, DbFlowTaskRun])

        await init_beanie(database=mongo_client[self.db_name], document_models=document_models)



class DummyLogger:
    """
    Logger thant do nothing
    """
    def debug(self, msg: str):
        ...

    def info(self, msg: str):
        ...

    def warning(self, msg: str):
        ...

    def error(self, msg: str):
        ...

    def exception(self, msg: str):
        ...


class DelayFlowTask(Exception):
    """
    Exception should be raised in task cb to delay its execution on
    `delay` seconds
    """
    def __init__(self, delay: int, *args: object) -> None:
        super().__init__(*args)
        self.delay = delay


class StopFlowRun(Exception):
    """ """



from typing import Protocol, Optional, TYPE_CHECKING

from aioredis import Redis

from kmx.models import MessageEnvelope, RpcQuery, RpcResult

from kmx.services.logger import Logger
from kmx.services.locker import Locker


if TYPE_CHECKING:
    from kmx.agents import Agent


class Container(Protocol):
    """ """

    logger: Logger
    locker: Locker
    redis: Optional[Redis]

    def add_agent(self, agent: 'Agent'):
        """ """

    async def start(self):
        """ """

    async def send_message(self, message_envelope: MessageEnvelope):
        """ """

    async def rpc_call(self, rpc_envelope: RpcQuery) -> RpcResult:
        """ """



from aiokafka import AIOKafkaProducer

from kmx.models import MessageEnvelope
from kmx.services.logger import Logger
from kmx.services.logger.dummy import DummyLogger


class KafkaMessageProducer:
    """ """
    def __init__(self, kafka_url: str, logger: Logger = None):
        self._kafka_url = kafka_url
        self._producer: AIOKafkaProducer = None  # noqa
        self._logger = logger or DummyLogger()

    async def start(self):
        if self._producer is not None:
            self._logger.info('Stopping current producer')
            await self._producer.stop()

        self._producer = AIOKafkaProducer(bootstrap_servers=self._kafka_url)
        self._logger.info(f'Kafka producer created for bootstrap_server `{self._kafka_url}`')
        await self._producer.start()

    async def stop(self):
        if self._producer is not None:
            await self._producer.stop()
            self._producer = None

    async def send_message(self, message: MessageEnvelope):
        """ """
        self._logger.debug(f'producing message: {message}')
        await self._producer.send_and_wait(topic=message.receiver, value=message.json(ensure_ascii=False).encode())



from typing import Any, Optional

import abc
import asyncio

from beanie.operators import And, Eq

from kmx.flows import Flow, FlowPack, FlowProcessor
from kmx.flows.db import DbFlowRun, DbFlowTaskRun, DbFlowRunState, DbFlowTaskRunState

from kmx.services.logger import Logger

from .base import BehaviourBase


class FlowBehaviour(BehaviourBase):
    """ """
    DbFlowRun = DbFlowRun
    DbFlowRunState = DbFlowTaskRunState

    def __init__(self, logger: Logger):
        super().__init__(logger=logger)
        self.flow_processor: Optional[FlowProcessor] = None
        self.flows: Optional[FlowPack] = None

    @abc.abstractmethod
    def create_flows(self) -> FlowPack:
        """ """

    async def run(self):
        """ 
        Старт поведения по работе с потоками
        """
        self.flows = self.create_flows()
        self.flow_processor = FlowProcessor(
            flows=self.flows,
            logger=self.logger,
            locker=self.locker()
        )

        self.logger.info(f'Added flows: {",".join(self.flows.flow_names)}')

        asyncio.create_task(self.flow_processor.scheduler())

    async def run_flow(self, flow: Flow, params: Any = None, delay: int = None):
        """ """
        await self.flow_processor.run_flow(
            flow=flow,
            params=params,
            delay=delay
        )

    async def get_flow_run_by_uid(self, flow_name: str, flow_run_uid: str) -> Optional[DbFlowRun]:
        """ """
        search_criteria = And(
            Eq(DbFlowRun.flow_name, flow_name),
            Eq(DbFlowRun.uid, flow_run_uid)
        )

        return await DbFlowRun.find_one(search_criteria)



from typing import Protocol

from kmx.models import MessageEnvelope


class MessageProducer(Protocol):
    """ """

    async def start(self):
        """ """

    async def stop(self):
        """ """

    async def send_message(self, message: MessageEnvelope):
        """ """



from typing import Optional, Any, Awaitable, Callable

import asyncio
import json
import inspect

from aioredis import Redis
from pydantic import BaseModel, ValidationError

from kmx.models import Message, MessageEnvelope, RpcQuery, RpcResult
from kmx.agents import Agent
from kmx.services.message_consumer import MessageConsumer
from kmx.services.message_producer import MessageProducer
from kmx.services.locker import Locker

from kmx.services.logger import Logger
from kmx.services.logger.dummy import DummyLogger

from kmx.db import DbInitializer

from .base import ContainerBase
from ..exceptions import RpcTimeout, RpcIncompatibleParams, RpcBadQuery


class KmxContainer(ContainerBase):
    """ """
    def __init__(self,
                 message_consumer: MessageConsumer,
                 message_producer: MessageProducer,
                 locker: Locker,
                 logger: Logger,
                 redis: Optional[Redis] = None,
                 db_initializer: Optional[DbInitializer] = None):
        """ """
        super().__init__(logger=logger, locker=locker, redis=redis)

        self.message_consumer = message_consumer
        self.message_producer = message_producer
        self.db_initializer = db_initializer

        self.rpc_map: dict[tuple[str, str], Callable] = {}

    async def start(self):
        """ """
        self.logger.info('Starting container')

        if not self.agents:
            self.logger.error('Container has no agents. Finishing')
            return

        # init db
        if self.db_initializer is not None:
            await self.db_initializer.init_db()
            self.logger.info('Database initialized')

        # start producer
        await self.message_producer.start()

        # create task list
        tasks = [
            asyncio.create_task(self.consume_messages())
        ]
        if self.redis is not None:
            self._parse_agents_rpc()
            tasks.append(self.listen_rpc_queue())

        for agent in self.agents:
            await agent.start()
            for behav in agent.behaviours:
                tasks.append(asyncio.create_task(behav.run()))

        await asyncio.gather(*tasks)

    async def send_message(self, message_envelope: MessageEnvelope):
        """ """
        await self.message_producer.send_message(message_envelope)

    # ------------------------------------------------------------------------------------------------------------------
    # Обработка входящих сообщений
    # ------------------------------------------------------------------------------------------------------------------
    async def consume_messages(self):
        topics = []
        for agent in self.agents:
            topics.extend(agent.topics)
        try:
            await self.message_consumer.start(topics=set(topics))
        except:
            self.logger.exception('Cannot start message consumer')

        try:
            async for message in self.message_consumer.consume_messages():
                self.logger.debug(f'Получено сообщение {message.json(ensure_ascii=False)}')
                agent = self.get_agent_by_name(message.receiver)
                if agent is not None:
                    await agent.receive_message(message)
        finally:
            await self.message_consumer.stop()

    # ------------------------------------------------------------------------------------------------------------------
    # Обработка RPC сообщений
    # ------------------------------------------------------------------------------------------------------------------
    async def listen_rpc_queue(self):
        """ """
        queue_names = [self._rpc_queue_name(agent.name) for agent in self.agents]

        self.logger.info(f'Start listening rpc queues `{queue_names}`')

        async with self.redis:

            while True:
                try:
                    queue_name, message = await self.redis.blpop(queue_names)
                    self.logger.debug(f'Got rpc call from queue {queue_name}: {message}')

                    #await self._handle_rpc_call(message)
                    asyncio.create_task(self._handle_rpc_call(message))

                except RpcBadQuery as e:
                    self.logger.error(f'Got rpc call with bad query: {e}')
                except Exception as e:
                    self.logger.exception('Error on handle rpc call')
                    await asyncio.sleep(0.05)

    async def _handle_rpc_call(self, message: bytes):
        """ """
        try:
            rpc_query: RpcQuery = RpcQuery.parse_raw(message)
        except ValidationError as e:
            rpc_call_uid = self._try_get_rpc_call_uid(message)
            if rpc_call_uid is None:
                # отправляем ответ с ошибкой
                raise RpcBadQuery(f'Cannot read rpc_call_uid from message {message}')
            else:
                error_msg = f'Rpc bad query: {e}'
                self.logger.error(f'Got rpc with bad query: {e}')
                await self._reply_rpc_error(rpc_call_uid, error_msg, timeout=5)
                return

        # Check rpc timeout
        if (await self.redis.delete(self._rpc_expire_key(rpc_query.rpc_call_uid))) == 0:
            # timeout - do nothing
            self.logger.warning(f'Got rpc call with timeout: {rpc_query}')
            return

        # get rpc_worker
        rpc_executor: Callable = self.rpc_map.get((rpc_query.rpc_to, rpc_query.rpc_name))
        if rpc_executor is None:
            error_msg = f'Rpc executor for {rpc_query.rpc_to}.{rpc_query.rpc_name} not found'
            self.logger.error(error_msg)

            await self._reply_rpc_error(rpc_query.rpc_call_uid, error_msg, rpc_query.timeout)
            return

        # execute rpc
        try:
            params_found, params = self._rpc_params_for_executor(rpc_executor, rpc_query.data)
        except ValidationError as e:
            error_msg = \
                f'Got incompatible params for rpc executor {rpc_query.rpc_to}.{rpc_query.rpc_name}: {e}'
            self.logger.error(error_msg)

            await self._reply_rpc_error(rpc_query.rpc_call_uid, error_msg, rpc_query.timeout)
            return

        try:
            if params_found:
                executor_result = await rpc_executor(params=params)
            else:
                executor_result = await rpc_executor(params)
        except Exception as e:
            error_msg = \
                f'Rpc remote error: {e}'
            self.logger.exception(f'Error on rpc {rpc_query.rpc_to}.{rpc_query.rpc_name} execution')
            await self._reply_rpc_error(rpc_query.rpc_call_uid, error_msg, rpc_query.timeout)
            return

        # всё ок, отправляем результат
        await self._reply_rpc_result(rpc_query, executor_result)

    async def _reply_rpc_result(self, rpc_query: RpcQuery, executor_result: Any):
        """ """
        result_key = self._rpc_results_key(rpc_call_uid=rpc_query.rpc_call_uid)

        if isinstance(executor_result, BaseModel):
            data = executor_result.dict()
        elif isinstance(executor_result, dict):
            data = executor_result
        else:
            # непонятное значение, попробуем передать как есть
            data = {'__rpc_raw_result__': executor_result}

        executor_result = RpcResult(rpc_call_uid=rpc_query.rpc_call_uid, data=data)

        # send result
        await self.redis.lpush(result_key, executor_result.json(ensure_ascii=False).encode())
        # set timeout for result
        await self.redis.expire(result_key, rpc_query.timeout)

        return

    async def _reply_rpc_error(self, rpc_call_uid: str, error: str, timeout: int):
        """ """
        result_key = self._rpc_results_key(rpc_call_uid)
        result = RpcResult(rpc_call_uid=rpc_call_uid, error=error)

        # send result
        await self.redis.lpush(result_key, result.json(ensure_ascii=False).encode())
        # set timeout for result
        await self.redis.expire(result_key, timeout)

        return

    @staticmethod
    def _try_get_rpc_call_uid(self, rpc_message: bytes) -> Optional[str]:
        """ """
        result = None

        try:
            data = json.loads(rpc_message)
            if isinstance(data, dict):
                result = data.get('rpc_call_uid')
        except:
            ...

        return result

    async def rpc_call(self, rpc_query: RpcQuery) -> RpcResult:
        """ """
        # set timeout for receiver
        await self.redis.setex(
            self._rpc_expire_key(rpc_query.rpc_call_uid),
            rpc_query.timeout,
            1  # any value
        )
        # send request
        await self.redis.rpush(
            self._rpc_queue_name(rpc_query.rpc_to),
            rpc_query.json(ensure_ascii=False).encode()
        )

        # wait for results (with timeout)
        async with self.redis:
            result_message = await self.redis.blpop(
                self._rpc_results_key(rpc_query.rpc_call_uid),
                rpc_query.timeout
            )

        if result_message is None:
            raise RpcTimeout(
                f'Timeout for rpc call `{rpc_query.rpc_to}.{rpc_query.rpc_name}` ({rpc_query.timeout} sec.)')

        _, message = result_message

        return RpcResult.parse_raw(message)

    @staticmethod
    def _rpc_queue_name(to: str) -> str:
        return f'kmx:{to}:rpc_queue'

    @staticmethod
    def _rpc_expire_key(rpc_call_uid: str) -> str:
        return f'kmx:rpc_expiry_key:{rpc_call_uid}'

    @staticmethod
    def _rpc_results_key(rpc_call_uid: str) -> str:
        """ """
        return f'kmx:rpc_result:{rpc_call_uid}'

    def _parse_agents_rpc(self):
        """ """
        self.rpc_map = {}
        for agent in self.agents:
            for _, func in inspect.getmembers(agent, predicate=inspect.iscoroutinefunction):
                if hasattr(func, '__kmx_rpc_function__'):
                    self.rpc_map[(agent.name, func.__kmx_rpc_function__)] = func
        return

    @staticmethod
    def _rpc_params_for_executor(executor: Any, params_dict: dict) -> tuple[bool, Any]:
        """ """
        params = params_dict

        arg_params = inspect.signature(executor).parameters.get('params')

        if arg_params is None:
            return False, params

        if issubclass(arg_params.annotation, BaseModel):
            params = arg_params.annotation.parse_obj(params)

        return True, params


# Compability
GenericContainer = KmxContainer



import asyncio
import time
from contextlib import asynccontextmanager

import aioredis
from aioredis.exceptions import RedisError

from kmx.services.logger import Logger
from kmx.services.logger.dummy import DummyLogger

from kmx.exceptions import LockTimeout


class RedisLocker:
    """ """
    def __init__(self,
                 redis_url: str,
                 logger: Logger | None = None,
                 tick: float = 0.5,
                 default_expires_in: int = 60,
                 prefix: str | None = None):
        """ """
        self._redis = aioredis.from_url(redis_url)
        self._logger = logger or DummyLogger()
        self._tick = tick
        self._default_expires_in = default_expires_in
        self._prefix = prefix or 'lock'

    @asynccontextmanager
    async def lock(self, identifier: str, expires_in: int | None = None, wait_time: int | None = None):
        """ """
        expires_in = expires_in if expires_in is not None else self._default_expires_in
        lock_info = time.time()
        async with self._wait(identifier, wait_time):
            try:
                await self._redis.setex(self._lock_id(identifier), expires_in, lock_info)
                yield
            finally:
                try:
                    await self._redis.delete(self._lock_id(identifier))
                except RedisError:
                    # ну и ладно
                    ...

    @asynccontextmanager
    async def wait(self, identifier: str, wait_time: int | None = None):
        """ """
        async with self._wait(identifier, wait_time):
            yield

    @asynccontextmanager
    async def _wait(self, identifier: str, wait_time: int | None = None):
        """ """
        started_at = time.monotonic()
        while True:
            if not await self._redis.exists(self._lock_id(identifier)):
                break
            await asyncio.sleep(self._tick)
            if not wait_time or (time.monotonic() - started_at) > wait_time:
                raise LockTimeout(f'Lock timeout for identifier {identifier}')
        yield

    def _lock_id(self, identifier: str) -> str:
        return f'{self._prefix}_{identifier}'



from typing import AsyncIterable, Any

import asyncio
import datetime
import uuid

from pydantic import BaseModel, ValidationError

from beanie.operators import And, In, Eq, Or, GTE, Set
import pymongo

from kmx.services.locker import Locker
from kmx.exceptions import LockTimeout

from kmx.services.logger import Logger
from kmx.services.logger.dummy import DummyLogger

from .flow import Flow, FlowPack
from .models import FlowRun
from .tasks import FlowTask
from .db import DbFlowRun, DbFlowRunState, DbFlowTaskRun, DbFlowTaskRunState, DbActiveFlowRunView, \
    DbFinishedFlowTaskRunView
from .exceptions import DelayFlowTask, StopFlowRun


class FlowProcessor:

    def __init__(self, flows: FlowPack, locker: Locker, logger: Logger | None = None):
        self.flows = flows
        self.locker = locker
        self.start_delay = 1
        self.delay = 0.05
        self.logger = logger or DummyLogger()

    async def run_flow(self, flow: Flow, params: Any = None, delay: int = None) -> DbFlowRun:
        """
        Run flow

        Args:
            flow: flow definition to run
            params: start params (dict or instance of BaseModel subclass)
            delay:
        """
        if not self.flows.flow_exists(flow):
            raise Exception(f'Flow `{flow.name}` not found')
        
        flow_run = DbFlowRun(
            flow_name=flow.name,
            uid=str(uuid.uuid4()).replace('-', ''),
            context={
                'flow_params': params.dict() if isinstance(params, BaseModel) else params
            },
            state=DbFlowRunState.ACTIVE,
            created_at=datetime.datetime.now(datetime.timezone.utc),
            delayed_until=datetime.datetime.now(datetime.timezone.utc) + datetime.timedelta(seconds=delay) if delay else None,
            finished_at=None
        )

        await flow_run.create()

        self.logger.info(f'Flow `{flow.name}` started with params {params} and delay {delay} sec')

        return flow_run

    async def scheduler(self, once: bool = False):
        """ """
        self.logger.info('Start flow scheduling')
        await asyncio.sleep(self.start_delay)
        while True:
            has_active_runs = False
            async for flow_run in self.fetch_active_flow_runs():
                has_active_runs = True
                if flow_run.delayed_until is not None and \
                        flow_run.delayed_until > datetime.datetime.now(datetime.timezone.utc):
                    continue
                try:
                    async with self.locker.lock(flow_run.uid):
                        await self.handle_active_flow_run(flow_run)
                except LockTimeout:
                    # поток уже заблокирован, выполняется где-то
                    continue
                except Exception as e:
                    raise
            if once and not has_active_runs:
                break
            await asyncio.sleep(self.delay)

    def fetch_active_flow_runs(self) -> AsyncIterable[DbActiveFlowRunView]:
        """ """
        filter_expr = And(
            In(DbFlowRun.flow_name, self.flows.flow_names),
            Eq(DbFlowRun.state, DbFlowRunState.ACTIVE),
            # Or(
            #     Eq(DbFlowRun.delayed_until, None),
            #     GTE(DbFlowRun.delayed_until, datetime.datetime.now(datetime.timezone.utc))
            # )
        )

        return DbFlowRun.find(filter_expr).sort([('created_at', pymongo.ASCENDING)]).project(DbActiveFlowRunView)

    async def handle_active_flow_run(self, flow_run_view: DbActiveFlowRunView):
        """ """
        # Получаем список уже выполненных задач
        done_task_names = []

        flow_run = await DbFlowRun.find_one(DbFlowRun.id == flow_run_view.id)
        if flow_run is None:
            # Flow run does not exist, skipping
            return

        async for task_run in DbFlowTaskRun\
                .find(DbFlowTaskRun.flow_run_uid == flow_run_view.uid).project(DbFinishedFlowTaskRunView):
            if task_run.state == DbFlowTaskRunState.ACTIVE:
                # There is an active task for the flow. Skip!
                return
            done_task_names.append(task_run.task_name)

        flow = self.flows.get_flow_by_name(flow_run_view.flow_name)
        tasks_to_run = self.get_tasks_to_run(flow, done_task_names)

        if not tasks_to_run:
            # There is no a task in the flow to run. Finalize it!
            await flow_run.update(Set({
                DbFlowRun.state: DbFlowRunState.FINISHED,
                DbFlowRun.finished_at: datetime.datetime.now(datetime.timezone.utc)  
            }))
            self.logger.info(
                f'Flow `{flow.name}` with params {flow_run.context.get("flow_params")} finished: all tasks completed')
        else:
            for flow_task in tasks_to_run:
                asyncio.create_task(self.run_task(flow, flow_run, flow_task))

    def get_tasks_to_run(self, flow: Flow, done_task_names: list[str]) -> list[FlowTask]:
        """ """
        if not done_task_names:
            return flow.heads()

        tasks = []
        for done_task_name in done_task_names:
            tasks.extend([t for t in flow.successors(done_task_name) if t.name not in done_task_names])

        return tasks

    async def run_task(self, flow: Flow, flow_run: DbFlowRun, flow_task: FlowTask):
        """ """
        # create task_run for flow_task
        task_run = DbFlowTaskRun(
            flow_run_uid=flow_run.uid,
            task_name=flow_task.name,
            state=DbFlowTaskRunState.ACTIVE,
            created_at=datetime.datetime.now(datetime.timezone.utc),
            finished_at=None
        )
        await task_run.insert()

        try:
            context = flow.context_model.parse_obj(
                flow_run.context) if flow.context_model is not None else flow_run.context
        except ValidationError as e:
            # Cannot parse context to desired model correctly. Stopping this flow
            await task_run.delete()

            context = flow_run.context
            context['_error'] = {
                'code': 'validation-error',
                'text': f'{e}'
            }

            await flow_run.update(Set({
                DbFlowRun.state: DbFlowTaskRunState.FINISHED,
                DbFlowRun.finished_at: datetime.datetime.now(datetime.timezone.utc),
                DbFlowRun.context: context
            }))
            self.logger.error(
                f'Flow `{flow.name}` with params {context.get("flow_params")} finished with '
                f'context validation error: {e}')
            return

        try:
            await flow_task.run(context, FlowRun(uid=flow_run.uid, flow_name=flow_run.flow_name))
        except StopFlowRun as e:
            context = context.dict() if isinstance(context, BaseModel) else context
            context['_reason'] = {
                'code': 'worflow-stopped',
                'text': f'Worflow was stopped: {e}'
            }
            await flow_run.update(Set({
                DbFlowRun.state: DbFlowTaskRunState.FINISHED,
                DbFlowRun.finished_at: datetime.datetime.now(datetime.timezone.utc),
                DbFlowRun.context: context
            }))
            await task_run.delete()
            self.logger.info(
                f'Flow `{flow.name}` with params {context.get("flow_params")} finished by command: {e}'
            )

        except DelayFlowTask as e:
            await flow_run.update(Set({
                DbFlowRun.delayed_until: datetime.datetime.now(datetime.timezone.utc) + datetime.timedelta(
                    seconds=e.delay),
                DbFlowRun.context: context.dict() if isinstance(context, BaseModel) else context
            }))
            await task_run.delete()

        except Exception as e:
            # получено непонятное сообщение, финализируем даг
            self.logger.exception(
                f'Исключение при попытке выполнения шага {task_run.task_name} потока {flow_run.flow_name}')
            context = context.dict() if isinstance(context, BaseModel) else context
            context['_reason'] = {
                'code': 'exception',
                'text': f'{e}'
            }
            await flow_run.update(Set({
                DbFlowRun.state: DbFlowTaskRunState.FINISHED,
                DbFlowRun.finished_at: datetime.datetime.now(datetime.timezone.utc),
                DbFlowRun.context: context
            }))
            await task_run.delete()
            self.logger.error(
                f'Flow `{flow.name}` with params {context.get("flow_params")} finished with exception: {e}'
            )

        else:
            # update flow context after task run
            await flow_run.update(Set({
                DbFlowRun.context: context.dict() if isinstance(context, BaseModel) else context
            }))

            # in any case mark task_run as finished
            await task_run.update(Set({
                DbFlowTaskRun.state: DbFlowTaskRunState.FINISHED,
                DbFlowTaskRun.finished_at: datetime.datetime.now(datetime.timezone.utc)
            }))



from typing import Callable, Optional
import abc

from kmx.models import MessageEnvelope
from kmx.agents import Agent
from kmx.services.locker import Locker
from kmx.services.logger import Logger
from kmx.services.logger.dummy import DummyLogger

from kmx.agents.base import AgentContext


class BehaviourBase(metaclass=abc.ABCMeta):
    """ """

    def __init__(self, logger: Logger, agent: Optional[Agent] = None):
        self.agent = agent or AgentContext.current_agent()
        self.logger = logger

        self.agent.add_behaviour(self)

    def locker(self) -> Locker:
        """ """
        return self.agent.locker()

    async def run(self):
        """ """

    async def internal_run(self):
        await self.run()

    async def match_message(self, message_envelope: MessageEnvelope) -> Optional[Callable]:
        return None



from typing import TYPE_CHECKING, Optional, Any

from collections import deque

from pydantic import BaseModel
import networkx as nx


if TYPE_CHECKING:
    from kmx.flows.tasks import FlowTask


class Flow:
    """ """
    def __init__(self, name: str, context_model: Any = None):
        self.name = name
        self.context_model = context_model
        self.tasks: list['FlowTask'] = []
        self.graph = nx.DiGraph()

    def __enter__(self) -> 'Flow':
        FlowDAGContext.push_flow(self)
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        FlowDAGContext.pop_flow()

    def get_task_by_name(self, task_name: str) -> Optional['FlowTask']:
        """ 
        Возвращает задачу по её названию
        """
        return next((t for t in self.tasks if t.name == task_name), None)

    def add_task(self, task: 'FlowTask') -> 'FlowTask':
        """ """
        if self.get_task_by_name(task.name) is None:
            self.tasks.append(task)
            self.graph.add_node(task.name)
        return task

    def add_edge(self, task_from: 'FlowTask', task_to: 'FlowTask'):
        """ """
        if self.get_task_by_name(task_from.name) is None:
            raise Exception(f"Задача {task_from.name} не добавлена в Flow")

        if self.get_task_by_name(task_to.name) is None:
            raise Exception(f"Задача {task_to.name} не добавлена в DAG {self.name}")

        self.graph.add_edge(task_from.name, task_to.name)

        return

    def heads(self) -> list['FlowTask']:
        """ """
        head_nodes = [node for node, degree in self.graph.in_degree if degree == 0]

        return [task for head_name in head_nodes if (task := self.get_task_by_name(head_name)) if not None]

    def successors(self, task_name: str) -> list['FlowTask']:
        """ """
        successor_names = self.graph.successors(task_name)

        return [task for name in successor_names if (task := self.get_task_by_name(name)) is not None]



class FlowDAGContext:
    """ """
    graphs = deque()

    @classmethod
    def push_flow(cls, graph: Flow):
        FlowDAGContext.graphs.append(graph)

    @classmethod
    def pop_flow(cls) -> Flow | None:
        return FlowDAGContext.graphs.pop() if FlowDAGContext.graphs else None

    @classmethod
    def current_flow(cls) -> Flow | None:
        return FlowDAGContext.graphs[-1] if FlowDAGContext.graphs else None


class FlowPack:
    """ """
    def __init__(self):
        """ """
        self.flows: dict[str, Flow] = {}

    def add_flow(self, flow: Flow) -> Flow:
        if flow.name in self.flows:
            raise Exception(f'Flow `{flow.name}` already added')
        self.flows[flow.name] = flow
        return flow

    @property
    def flow_names(self) -> list[str]:
        return [flow.name for _, flow in self.flows.items()]

    def flow_exists(self, flow: Flow) -> bool:
        return flow.name in self.flows.keys()

    def get_flow_by_name(self, flow_name: str) -> Flow:
        return self.flows[flow_name]

    def __getitem__(self, item: str) -> Flow:
        if item not in self.flows:
            raise Exception('Flow `name` not found in flowpack')
        return self.flows[item]



from typing import Optional
import asyncio

from kmx.services.logger import Logger

from .base import BehaviourBase


class PeriodicBehaviour(BehaviourBase):
    """ """

    def __init__(self, logger: Logger, period: int, start_delay: int = -1):
        """ """
        super().__init__(logger=logger)
        self.period = period
        self.start_delay = start_delay if start_delay >= 0 else period

    async def internal_run(self):
        """ """
        # стартовая пауза
        await asyncio.sleep(self.start_delay)

        while True:
            try:
                await self.run()
            except Exception as e:
                self.logger.exception(f'Ошибка запуска периодического задания у {self}')
            # Пауза между тиками
            await asyncio.sleep(self.period)

    async def run(self):
        """ """
        ...


from typing import Optional

import abc
from collections import deque

from aioredis import Redis

from kmx.models import MessageEnvelope
from kmx.containers import Container
from kmx.agents import Agent

from kmx.services.logger import Logger
from kmx.services.locker import Locker


class ContainerContext:
    """ """
    containers = deque()

    @classmethod
    def push_container(cls, container: Container):
        """ """
        ContainerContext.containers.append(container)

    @classmethod
    def pop_container(cls) -> Optional[Container]:
        """ """
        return ContainerContext.containers.pop() if ContainerContext.containers else None

    @classmethod
    def current_container(cls) -> Optional[Container]:
        """ """
        return ContainerContext.containers[-1] if ContainerContext.containers else None


class ContainerBase(metaclass=abc.ABCMeta):
    """ """
    def __init__(self, locker: Locker, logger: Logger, redis: Optional[Redis] = None):
        """ """
        self.agents: list[Agent] = []
        self.agents_map: dict[str, Agent] = {}
        self.locker = locker
        self.logger = logger
        self.redis = redis

    def __enter__(self) -> Container:
        """ """
        ContainerContext.push_container(self)
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        """ """
        ContainerContext.pop_container()

    def add_agent(self, agent: Agent):
        """ """
        if not agent.name:
            raise Exception(f'Не указано имя агента при добавлении в контейнер: {type(agent)}')
        if agent.name in self.agents_map:
            raise Exception(f'Agent with name `{agent.name}` already exists in the container')
        self.agents_map[agent.name] = agent

        self.agents.append(agent)

        self.logger.info(f'Agent `{agent.name}` added to container')

    @abc.abstractmethod
    async def start(self):
        """ """
        ...

    @abc.abstractmethod
    async def send_message(self, message_envelope: MessageEnvelope):
        """ """
        ...

    def get_agent_by_name(self, name: str) -> Optional[Agent]:
        """ """
        return self.agents_map.get(name)



from typing import Protocol
from collections.abc import AsyncIterator

from kmx.models import MessageEnvelope


class MessageConsumer(Protocol):
    """ """

    async def start(self, topics: list[str] | set[str] | tuple[str]):
        ...

    async def stop(self):
        ...

    async def consume_messages(self) -> AsyncIterator[MessageEnvelope]:
        """ """
        ...



import asyncio
import uuid
import inspect
from typing import Optional, Any, Callable

from pydantic import BaseModel

from kmx.models import Message, MessageEnvelope, RpcQuery
from kmx.containers import Container
from kmx.agents.base import AgentBase
from kmx.exceptions import RpcRemoteError
from kmx.services.locker import Locker

from kmx.services.logger import Logger


class KmxAgent(AgentBase):
    """ """

    def __init__(self, name: str, logger: Logger, topics: Optional[list[str]] = None,
                 container: Optional[Container] = None):
        super().__init__(name=name, logger=logger, topics=topics, container=container)
        self.message_handlers_map: dict[str, list[Callable]] = {}

    async def start(self):
        """ """
        self._parse_message_handlers()

    def locker(self) -> Locker:
        return self.container.locker

    # ------------------------------------------------------------------------------------------------------------------
    # Работа с сообщениями
    # ------------------------------------------------------------------------------------------------------------------
    async def send_message(self, to: str, message: Message, reply_to: Optional[Message] = None):
        """ """
        message_envelope = MessageEnvelope(
            sender=self.name,
            receiver=to,
            message=message,
            reply_to=reply_to
        )
        await self.container.send_message(message_envelope)

    async def rpc_call(self, to: str, what: str, params: Any, response_class: Optional[Any] = None) -> Any:
        """ """
        rpc_envelope = RpcQuery(
            rpc_call_uid=str(uuid.uuid4()).replace('-', ''),
            rpc_to=to,
            rpc_name=what,
            data=params.dict() if isinstance(params, BaseModel) else params
        )

        result = await self.container.rpc_call(rpc_envelope)

        if result.error is not None:
            raise RpcRemoteError(result.error)

        return response_class.parse_obj(result.data) if issubclass(response_class, BaseModel) else result.data

    async def receive_message(self, message_envelope: MessageEnvelope):
        """ """
        # составляем список обработчиков сообщения
        handlers = []
        if message_envelope.message.code in self.message_handlers_map:
            handlers.extend(self.message_handlers_map[message_envelope.message.code])

        for behav in self.behaviours:
            handler = await behav.match_message(message_envelope)

            if handler is not None:
                handlers.append(handler)

        if not handlers:
            self.logger.warning(f'Для сообщения с кодом {message_envelope.message.code} нет обработчика у агента '
                                f'{self.name}')

        for handler in handlers:
            asyncio.create_task(self._handle_message(handler, message_envelope))

    async def _handle_message(self, message_handler: Callable, message_envelope: MessageEnvelope):
        """ """
        try:
            kwargs = self._prepare_message_handler_params(message_handler, message_envelope)
            await message_handler(**kwargs)
        except Exception as e:
            self.logger.exception(f'Ошибка выполнения обработчика сообщения {message_envelope.message.data}')

    def _parse_message_handlers(self):
        """ """
        self.message_handlers_map = {}
        for behaviour in self.behaviours:
            for _, func in inspect.getmembers(behaviour, predicate=inspect.iscoroutinefunction):
                if hasattr(func, '__kmx_message_code__'):
                    self.message_handlers_map.setdefault(func.__kmx_message_code__, []).append(func)
        return

    @staticmethod
    def _prepare_message_handler_params(message_handler: Any, message_envelope: MessageEnvelope) -> dict:
        """ """
        handler_params = {}

        arg_message_envelope = inspect.signature(message_handler).parameters.get('message_envelope')
        arg_message_body = inspect.signature(message_handler).parameters.get('message_body')
        arg_reply_to_body = inspect.signature(message_handler).parameters.get('reply_to_body')

        if arg_message_envelope is not None:
            handler_params['message_envelope'] = message_envelope

        if arg_message_body is not None:
            if issubclass(arg_message_body.annotation, BaseModel):
                handler_params['message_body'] = arg_message_body.annotation.parse_obj(message_envelope.message.data)
            else:
                handler_params['message_body'] = arg_message_body

        if arg_reply_to_body is not None:
            if message_envelope.reply_to is None:
                handler_params['reply_to_body'] = None

            elif issubclass(arg_reply_to_body.annotation, BaseModel):
                handler_params['reply_to_body'] = arg_message_body.annotation.parse_obj(message_envelope.reply_to.data)

            else:
                handler_params['reply_to_body'] = arg_message_body

        return handler_params


GenericAgent = KmxAgent



from typing import Protocol, Optional, Any, TYPE_CHECKING

from kmx.models import Message, MessageEnvelope
from kmx.containers import Container
from kmx.services.locker import Locker
from kmx.services.logger import Logger


if TYPE_CHECKING:
    from kmx.behaviours import Behaviour


class Agent(Protocol):
    """ """

    name: str
    topics: list[str]
    container: Container
    behaviours: list['Behaviour']
    logger: Logger

    def add_behaviour(self, benaviour: 'Behaviour'):
        """ """
        ...

    # ------------------------------------------------------------------------------------------------------------------
    # Инициализация
    # ------------------------------------------------------------------------------------------------------------------
    async def start(self):
        """ """

    # ------------------------------------------------------------------------------------------------------------------
    # Работа с сообщениями
    # ------------------------------------------------------------------------------------------------------------------
    async def send_message(self, to: str, message: Message, reply_to: Optional[Message] = None):
        """ """

    async def receive_message(self, message_envelope: MessageEnvelope):
        """ """

    async def rpc_call(self, to: str, what: str, params: Any, response_class: Optional[Any]) -> Any:
        """ """

    def locker(self) -> Locker:
        """ """



from collections.abc import AsyncIterator

import asyncio
from aiokafka import AIOKafkaConsumer

from kmx.models import MessageEnvelope

from kmx.services.logger import Logger
from kmx.services.logger.dummy import DummyLogger


class KafkaMessageConsumer:
    """ """

    def __init__(self, kafka_url: str, group_id: str, logger: Logger = None):
        """ """
        self._kafka_url = kafka_url
        self._group_id = group_id
        self._consumer: AIOKafkaConsumer = None  # noqa
        self._logger = logger or DummyLogger()

    async def start(self, topics: list[str]):
        """ """
        if self._consumer is not None:
            self._logger.info('Stopping current consumer')
            await self._consumer.stop()

        self._consumer = AIOKafkaConsumer(
            *topics,
            bootstrap_servers=self._kafka_url,
            group_id=self._group_id,
            auto_offset_reset='earliest'
        )
        self._logger.info(f'Kafka consumer created for bootstrap_servers `{self._kafka_url}` and topics `{", ".join(topics)}`')
        await self._consumer.start()

    async def stop(self):
        if self._consumer is not None:
            await self._consumer.stop()
            self._consumer = None

    async def consume_messages(self) -> AsyncIterator[MessageEnvelope]:
        """ """
        self._logger.info('Start consuming messages')
        async for message in self._consumer:
            self._logger.debug(f'Consumed message `{message.value}`')
            try:
                message_envelope = MessageEnvelope.parse_raw(message.value)
            except:
                self._logger.exception(f'Cannot parse message envelope from `{message.value}`')
            else:
                await self._consumer.commit()
                yield message_envelope



from typing import Protocol, Callable

import abc
import inspect

from pydantic import BaseModel

from .models import FlowRun
from .flow import Flow, FlowDAGContext


# -------------------------------------------------------------------------------------------------
#
# -------------------------------------------------------------------------------------------------
class FlowTask(Protocol):
    """ 
    Определение задачи в потоке 
    """

    name: str
    flow: Flow | None

    def __rshift__(self, other: 'FlowTask') -> 'FlowTask':
        """ """
        ...

    async def run(self, context: BaseModel | dict, flow_run: FlowRun):
        """ """


# -------------------------------------------------------------------------------------------------
# Базовый класс для задач
# ----------------------------------------------------------------------------------------------------------------------
class FlowTaskBase:
    """ """
    def __init__(self, name: str, flow: Flow | None = None):
        """ """
        if not name:
            raise Exception('Empty task name is not allowed')
        
        self.name = name
        self.flow = flow or FlowDAGContext.current_flow()

        if self.flow is None:
            raise Exception('No flow in context')

        self.flow.add_task(self)

    def __rshift__(self, other: FlowTask) -> FlowTask:
        """
        Add dependency between tasks in flow
        """
        if self.flow is None:
            raise Exception(f'Task `{self.name}` has no assigned flow')
        
        if other.flow is None:
            raise Exception(f'Task `{other.name}` has no assigned flow')
        
        if self.flow != other.flow:
            raise Exception(f"DAG for task {self.name} is not the same as for {other.name} "
                            f"(DAGs: {self.flow.name} and {other.flow.name})")
        
        self.flow.add_edge(self, other)
        
        return other

    @abc.abstractmethod
    async def run(self, context: BaseModel | dict, flow_run: FlowRun):
        """ """


class PythonTask(FlowTaskBase):
    """ 
    Задача, которая выполняет определенный питоновский код
    """
    def __init__(self, name: str, cb: Callable, flow: Flow | None = None):
        super().__init__(name=name, flow=flow)
        self.cb = cb

    async def run(self, context: BaseModel | dict, flow_run: FlowRun):
        cb_args = {}

        cb_spec = inspect.getfullargspec(self.cb)
        if 'context' in cb_spec.args:
            cb_args['context'] = context
        if 'flow_run' in cb_spec.args:
            cb_args['flow_run'] = flow_run

        await self.cb(**cb_args)



from typing import Protocol


class Logger(Protocol):
    """
    Logger protocol
    """

    def debug(self, msg: str):
        ...

    def info(self, msg: str):
        ...

    def warning(self, msg: str):
        ...

    def error(self, msg: str):
        ...

    def exception(self, msg: str):
        ...



from typing import Optional, Any

import json
import uuid

from aioredis import Redis
from pydantic import BaseModel

from kmx.exceptions import RpcTimeout, RpcRemoteError, RpcError
from kmx.models import RpcResult


class RedisRpcCaller:
    """ """
    def __init__(self, redis: Redis, timeout: Optional[int] = None):
        self.redis = redis
        self.timeout = timeout if timeout > 0 else 5

    async def rpc_call(self, rpc_to: str, rpc_name: str, rpc_timeout: Optional[int] = None,
                       rpc_response_model: Optional[Any] = None, **params) -> Optional[Any]:
        """ """
        rpc_call_uid = str(uuid.uuid4()).replace('-', '')
        expire_key = f'kmx:rpc_expiry_key:{rpc_call_uid}'
        queue_name = f'kmx:{rpc_to}:rpc_queue'
        results_key = f'kmx:rpc_result:{rpc_call_uid}'
        timeout = rpc_timeout or self.timeout

        rpc_query = {
            'rpc_call_uid': rpc_call_uid,
            'rpc_to': rpc_to,
            'rpc_name': rpc_name,
            'data': self.prepare_rpc_params(**params)
        }

        await self.redis.setex(expire_key, timeout, 1)
        await self.redis.rpush(queue_name, json.dumps(rpc_query).encode())

        async with self.redis:
            result_message = await self.redis.blpop(results_key, timeout)

        if result_message is None:
            raise RpcTimeout(f'Timeout for {rpc_to}.{rpc_name} ({timeout} sec)')

        _, message = result_message

        try:
            result = RpcResult.parse_raw(message)
        except Exception as e:
            raise RpcError(f'Unknown rpc result: {message} (expected RpcResult value)')

        if result.error:
            raise RpcRemoteError(result.error)

        return self.parse_rpc_result(result, rpc_response_model)

    @staticmethod
    def prepare_rpc_params(**kwargs) -> dict:
        """ """
        if len(kwargs) == 1 and 'params' in kwargs:
            # поддержка устаревшего вызова, когда в параметрах передается одно значение "params"
            params = kwargs['params']
            return params.dict() if isinstance(params, BaseModel) else params

        params = {}
        for key, value in kwargs.items():
            params[key] = value.dict() if isinstance(value, BaseModel) else value

        return params

    @staticmethod
    def parse_rpc_result(rpc_worker_result: RpcResult, rpc_response_model: Any):
        """ """
        if rpc_response_model is not None and issubclass(rpc_response_model, BaseModel):
            return rpc_response_model.parse_obj(rpc_worker_result.data)

        if '__rpc_raw_result__' in rpc_worker_result.data:
            return rpc_worker_result.data['__rpc_raw_result__']

        return rpc_worker_result.data



import functools


def message_handler(message_code: str):
    """ Decorator for message handler """
    def wrapper(func):
        func.__kmx_message_code__ = message_code

        @functools.wraps(func)
        async def wrapped(*args, **kwargs):
            return await func(*args, **kwargs)
        return wrapped

    return wrapper



import logging


class LoggingLogger:
    """
    Logging throw python "logging" module
    """
    def __init__(self, logger_name: str):
        self.logger = logging.getLogger(logger_name)

    def debug(self, text: str):
        self.logger.debug(text)

    def info(self, text: str):
        self.logger.info(text)

    def warning(self, text: str):
        self.logger.warning(text)

    def error(self, text: str):
        self.logger.error(text)

    def exception(self, text: str):
        self.logger.exception(text)



from typing import AsyncIterator, Callable, Coroutine, Any
import inspect

import asyncio

from kmx.models import MessageEnvelope


class DummyMessageConsumer:
    """ """

    def __init__(self,
                 request_messages_cb: Callable[[], list[MessageEnvelope]] | Coroutine[Any, Any, list[MessageEnvelope]] = None,
                 tick_delay: float = 0.5):
        """ """
        self._request_messages_cb = request_messages_cb
        self._tick_delay = tick_delay

    async def start(self):
        """ """

    async def stop(self):
        """ """

    async def consume_messages(self) -> AsyncIterator[MessageEnvelope]:
        """ """
        while True:
            await asyncio.sleep(self._tick_delay)
            messages = await self._request_new_messages()
            for message in messages:
                yield message

    async def _request_new_messages(self) -> list[MessageEnvelope]:
        """ """
        if self._request_messages_cb is None:
            return []

        if inspect.iscoroutinefunction(self._request_messages_cb):
            return await self._request_messages_cb()  # noqa

        return self._request_messages_cb()



from pydantic import BaseModel


class Message(BaseModel):
    """ """
    code: str
    data: dict


class MessageEnvelope(BaseModel):
    """ """
    sender: str
    receiver: str
    message: Message
    reply_to: Message | None



from typing import Optional
from pydantic import BaseModel, root_validator


class RpcQuery(BaseModel):
    """ """
    rpc_call_uid: str
    rpc_to: str
    rpc_name: str
    data: dict
    timeout: int = 5


class RpcResult(BaseModel):
    """ """

    rpc_call_uid: str
    data: Optional[dict] = None
    error: Optional[str] = None

    @root_validator
    def check_data_and_error(cls, values):
        """ """
        if values['data'] is not None and values['error'] is not None:
            raise ValueError('Both `data` and `error` cannot be not None')
        return values



from typing import Protocol, Any, Optional


class RpcCaller(Protocol):
    """ """
    async def rpc_call(self, rpc_to: str, rpc_name: str, rpc_timeout: Optional[int] = None,
                       rpc_response_model: Optional[Any] = None, **params) -> Optional[Any]:
        """ """


class RpcWorker(Protocol):
    """ ДОделать """



import functools


def rpc(name: str):
    """ Decorator for API functions """
    def wrapper(func):
        func.__kmx_rpc_function__ = name

        @functools.wraps(func)
        async def wrapped(*args, **kwargs):
            return await func(*args, **kwargs)
        return wrapped

    return wrapper



from typing import Optional, Any

import abc
from collections import deque

from kmx.models import MessageEnvelope, Message
from kmx.agents import Agent
from kmx.containers import Container
from kmx.containers.base import ContainerContext
from kmx.behaviours import Behaviour
from kmx.services.logger import Logger


class AgentContext:
    """ """
    agents = deque()

    @classmethod
    def push_agent(cls, agent: Agent):
        """ """
        AgentContext.agents.append(agent)

    @classmethod
    def pop_agent(cls) -> Optional[Agent]:
        """ """
        return AgentContext.agents.pop() if AgentContext.agents else None

    @classmethod
    def current_agent(cls) -> Optional[Agent]:
        """ """
        return AgentContext.agents[-1] if AgentContext.agents else None


class AgentBase(metaclass=abc.ABCMeta):
    """ """

    def __init__(self, name: str, logger: Logger, topics: Optional[list[str]] = None,
                 container: Optional[Container] = None):
        """ """
        self.name = name
        self.topics = topics or [name]
        self.container = container or ContainerContext.current_container()
        self.behaviours: list[Behaviour] = []
        self.logger = logger

        self.container.add_agent(self)

    def __enter__(self) -> Agent:
        """ """
        AgentContext.push_agent(self)
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        """ """
        AgentContext.pop_agent()

    def add_behaviour(self, behaviour: Behaviour):
        self.behaviours.append(behaviour)

        self.logger.info(f'Behaviour `{behaviour.__class__.__name__}` added to agent `{self.name}`')

    @abc.abstractmethod
    async def start(self):
        """ """

    @abc.abstractmethod
    async def send_message(self, to: str, message: Message, reply_to: Optional[Message] = None):
        """ """

    @abc.abstractmethod
    async def receive_message(self, message_envelope: MessageEnvelope):
        """ """

    @abc.abstractmethod
    async def rpc_call(self, to: str, what: str, params: Any, response_class: Optional[Any]) -> Any:
        """ """




from typing import Protocol

from contextlib import asynccontextmanager


class Locker(Protocol):
    """ """

    @asynccontextmanager
    async def lock(self, identifier: str, expires_in: int | None = None, wait_time: int | None = None):
        """ """
        yield
        

    @asynccontextmanager
    async def wait(self, identifier: str, wait_time: int | None = None):
        """ """
        yield
        



#!/usr/bin/env python

from setuptools import setup, find_packages


setup(
    name='kmx',
    version='0.1.0',
    author='akvarats',

    classifiers=[
        'Programming Language :: Python :: 3',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent'
    ],
    packages=find_packages(),
    install_requires=[
        'aiokafka>=0.8,<0.9',
        'pydantic>=1.10,<1.11',
        'beanie>=1.17,<1.18',
        'aioredis[hiredis]>=2.0,<2.1',
        'networkx>=3.0,<3.1'
    ],
    setup_requires=['wheel'],
    python_requires='>=3.10,<4'
)



import datetime
from enum import IntEnum

from pydantic import BaseModel, Field
from beanie import Document, Link, PydanticObjectId
import pymongo


class DbFlowRunState(IntEnum):
    """ """
    ACTIVE = 1
    FINISHED = 2


class DbFlowRun(Document):
    """ """
    uid: str  # уникальный идентификатор запуска

    flow_name: str
    state: DbFlowRunState

    context: dict = {}  # контекст выполнения потока

    delayed_until: datetime.datetime | None = None

    created_at: datetime.datetime
    finished_at: datetime.datetime | None = None

    class Settings:
        name = 'flow_runs'


class DbActiveFlowRunView(BaseModel):
    """ """
    id: PydanticObjectId = Field(alias="_id")
    uid: str
    flow_name: str
    delayed_until: datetime.datetime | None


class DbFlowTaskRunState(IntEnum):
    """ """
    ACTIVE = 1
    FINISHED = 2


class DbFlowTaskRun(Document):
    """ """

    flow_run_uid: str
    task_name: str

    state: DbFlowTaskRunState

    created_at: datetime.datetime
    finished_at: datetime.datetime | None

    class Settings:
        name = 'flow_task_runs'
        indexes = [
            [('flow_run_uid', pymongo.ASCENDING)],
        ]


class DbFinishedFlowTaskRunView(BaseModel):
    """ Проекция документа DbFlowTaskRun для запросов завершившихся задач """
    id: PydanticObjectId = Field(alias='_id')
    task_name: str
    state: DbFlowTaskRunState



from typing import Callable, Optional

from kmx.models import MessageEnvelope

from .base import BehaviourBase


class SimpleBehaviour(BehaviourBase):
    """ """

    async def run(self):
        """ """
        # по умолчанию ничего не делаем

    async def match_message(self, message_envelope: MessageEnvelope) -> Optional[Callable]:
        """ """
        return None



from typing import Protocol, Callable, Optional

from kmx.models import MessageEnvelope
from kmx.agents import Agent
from kmx.services.locker import Locker
from kmx.services.logger import Logger


class Behaviour(Protocol):
    """ """
    agent: Agent
    logger: Logger

    async def run(self):
        """ """

    async def match_message(self, message_envelope: MessageEnvelope) -> Optional[Callable]:
        """ """

    def locker(self) -> Locker:
        """ """
